﻿//****************************************************************
//
// Gimp XCF Content Pipeline Importer
//
// Written by: Louis Ingenthron
// Last Modified: 28.05.2009
//
//****************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace XCFProcessorLib
{
    /// <summary>
    /// Represents a color using Hue, Saturation, Value, and Alpha values.
    /// </summary>
    public class ColorHSV
    {
        /// <summary>
        /// Gets or sets the hue component value.
        /// </summary>
        public byte H { get; private set; }
        /// <summary>
        /// Gets or sets the saturation component value.
        /// </summary>
        public byte S { get; private set; }
        /// <summary>
        /// Gets or sets the value component value.
        /// </summary>
        public byte V { get; private set; }
        /// <summary>
        /// Gets or sets the alpha component value.
        /// </summary>
        public byte A { get; private set; }

        /// <summary>
        /// Initializes a new instance of ColorHSV.
        /// </summary>
        /// <param name="colorRGB">The RGB Color to convert from</param>
        public ColorHSV(Color colorRGB)
        {
	        Vector3 hsv;
            Vector4 color = colorRGB.ToVector4();
	        float mn, mx, delta;
	        mn = Math.Min( color.X, Math.Min(color.Y, color.Z) );
	        mx = Math.Max( color.X, Math.Max(color.Y, color.Z) );
	        hsv.Z = mx;
	        delta = mx - mn;
	        if( mx > 0 )
	        {
		        hsv.Y = delta / mx;
		        if( color.X == mx )
			        hsv.X = ( color.Y - color.Z ) / delta;
		        else if( color.Y == mx )
			        hsv.X = 2 + ( color.Z - color.X ) / delta;
		        else
			        hsv.X = 4 + ( color.X - color.Y ) / delta;
		        hsv.X /= 6.0f;
		        if( hsv.X < 0 )
			        hsv.X += 1.0f;	
	        }
	        else 
	        {
		        hsv.Y = 0;
		        hsv.X = 0;
	        }
            H = (byte)(hsv.X * 255f);
            S = (byte)(hsv.Y * 255f);
            V = (byte)(hsv.Z * 255f);
            A = colorRGB.A;
        }

        /// <summary>
        /// Initializes a new instance of ColorHSV.
        /// </summary>
        /// <param name="colorRGB">The RGB Color to convert from</param>
        /// <param name="a">The Alpha component (0-255)</param>
        public ColorHSV(Color colorRGB, byte a)
        {
            Vector3 hsv;
            Vector4 color = colorRGB.ToVector4();
            float mn, mx, delta;
            mn = Math.Min(color.X, Math.Min(color.Y, color.Z));
            mx = Math.Max(color.X, Math.Min(color.Y, color.Z));
            hsv.Z = mx;
            delta = mx - mn;
            if (mx > 0)
            {
                hsv.Y = delta / mx;
                if (color.X == mx)
                    hsv.X = (color.Y - color.Z) / delta;
                else if (color.Y == mx)
                    hsv.X = 2 + (color.Z - color.X) / delta;
                else
                    hsv.X = 4 + (color.X - color.Y) / delta;
                hsv.X /= 6.0f;
                if (hsv.X < 0)
                    hsv.X += 1.0f;
            }
            else
            {
                hsv.Y = 0;
                hsv.X = 0;
            }
            H = (byte)(hsv.X * 255f);
            S = (byte)(hsv.Y * 255f);
            V = (byte)(hsv.Z * 255f);
            A = a;
        }

        /// <summary>
        /// Initializes a new instance of ColorHSV.
        /// </summary>
        /// <param name="colorRGB">The RGB Color to convert from</param>
        /// <param name="a">The Alpha component (0.0-1.0)</param>
        public ColorHSV(Color colorRGB, float a)
        {
            Vector3 hsv;
            Vector4 color = colorRGB.ToVector4();
            float mn, mx, delta;
            mn = Math.Min(color.X, Math.Min(color.Y, color.Z));
            mx = Math.Max(color.X, Math.Min(color.Y, color.Z));
            hsv.Z = mx;
            delta = mx - mn;
            if (mx > 0)
            {
                hsv.Y = delta / mx;
                if (color.X == mx)
                    hsv.X = (color.Y - color.Z) / delta;
                else if (color.Y == mx)
                    hsv.X = 2 + (color.Z - color.X) / delta;
                else
                    hsv.X = 4 + (color.X - color.Y) / delta;
                hsv.X /= 6.0f;
                if (hsv.X < 0)
                    hsv.X += 1.0f;
            }
            else
            {
                hsv.Y = 0;
                hsv.X = 0;
            }
            H = (byte)(hsv.X * 255f);
            S = (byte)(hsv.Y * 255f);
            V = (byte)(hsv.Z * 255f);
            A = (byte)(a * 255f);
        }

        /// <summary>
        /// Initializes a new instance of ColorHSV.
        /// </summary>
        /// <param name="h">The Hue component (0.0-1.0)</param>
        /// <param name="s">The Saturation component (0.0-1.0)</param>
        /// <param name="v">The Value component (0.0-1.0)</param>
        public ColorHSV(float h, float s, float v)
        {
            H = (byte)(h * 255f);
            S = (byte)(s * 255f);
            V = (byte)(v * 255f);
            A = 255;
        }

        /// <summary>
        /// Initializes a new instance of ColorHSV.
        /// </summary>
        /// <param name="h">The Hue component (0.0-1.0)</param>
        /// <param name="s">The Saturation component (0.0-1.0)</param>
        /// <param name="v">The Value component (0.0-1.0)</param>
        /// <param name="a">The Alpha component (0.0-1.0)</param>
        public ColorHSV(float h, float s, float v, float a)
        {
            H = (byte)(h * 255f);
            S = (byte)(s * 255f);
            V = (byte)(v * 255f);
            A = (byte)(a * 255f);
        }

        /// <summary>
        /// Initializes a new instance of ColorHSV.
        /// </summary>
        /// <param name="h">The Hue component (0-255)</param>
        /// <param name="s">The Saturation component (0-255)</param>
        /// <param name="v">The Value component (0-255)</param>
        public ColorHSV(byte h, byte s, byte v)
        {
            H = h;
            S = s;
            V = v;
            A = 255;
        }

        /// <summary>
        /// Initializes a new instance of ColorHSV.
        /// </summary>
        /// <param name="h">The Hue component (0-255)</param>
        /// <param name="s">The Saturation component (0-255)</param>
        /// <param name="v">The Value component (0-255)</param>
        /// <param name="a">The Alpha component (0-255)</param>
        public ColorHSV(byte h, byte s, byte v, byte a)
        {
            H = h;
            S = s;
            V = v;
            A = a;
        }

        /// <summary>
        /// Initializes a new instance of ColorHSV.
        /// </summary>
        /// <param name="col">The base HSV color to apply alpha to</param>
        /// <param name="a">The Alpha component (0-255)</param>
        public ColorHSV(ColorHSV col, byte a)
        {
            H = col.H;
            S = col.S;
            V = col.V;
            A = a;
        }

        /// <summary>
        /// Initializes a new instance of ColorHSV.
        /// </summary>
        /// <param name="col">The base HSV color to apply alpha to</param>
        /// <param name="a">The Alpha component (0.0-1.0)</param>
        public ColorHSV(ColorHSV col, float a)
        {
            H = col.H;
            S = col.S;
            V = col.V;
            A = (byte)(a*255f);
        }

        /// <summary>
        /// Initializes a new instance of ColorHSV.
        /// </summary>
        /// <param name="vector">A Framework.Vector3 containing the Hue, Saturation, and Value values defining a color.</param>
        public ColorHSV(Vector3 vector) : this(vector.X, vector.Y, vector.Z) { }

        /// <summary>
        /// Initializes a new instance of ColorHSV.
        /// </summary>
        /// <param name="vector">A Framework.Vector3 containing the Hue, Saturation, and Value values defining a color.</param>
        /// <param name="a">The Alpha component (0.0-1.0)</param>
        public ColorHSV(Vector3 vector, float a) : this(vector.X, vector.Y, vector.Z, a) { }

        /// <summary>
        /// Initializes a new instance of ColorHSV.
        /// </summary>
        /// <param name="vector">A Framework.Vector3 containing the Hue, Saturation, Value, and Alpha values defining a color.</param>
        public ColorHSV(Vector4 vector) : this(vector.X, vector.Y, vector.Z, vector.W) { }

        /// <summary>
        /// Returns the current color as a Framework.Vector3.
        /// </summary>
        /// <returns>The current color represented as a Framework.Vector3.</returns>
        public Vector3 ToVector3()
        {
            return new Vector3(H / 255f, S / 255f, V / 255f);
        }

        /// <summary>
        /// Returns the current color as a Framework.Vector4.
        /// </summary>
        /// <returns>The current color represented as a Framework.Vector4.</returns>
        public Vector4 ToVector4()
        {
            return new Vector4(H / 255f, S / 255f, V / 255f, A / 255f);
        }

        /// <summary>
        /// Returns an RGB Framework.Color representation of the ColorHSV
        /// </summary>
        /// <returns>A Framework.Color RGB equivalent color</returns>
        public Color ToRGB()
        {
            Vector3 hsv = this.ToVector3();
            Vector3 rgb;
            int i;
            float f, p, q, t;
            if (hsv.Y == 0)
            {
                rgb.X = hsv.Z;
                rgb.Y = hsv.Z;
                rgb.Z = hsv.Z;
            }
            else
            {
                hsv.X *= 6;
                i = (int)hsv.X;
                f = hsv.X - i;
                p = hsv.Z * (1 - hsv.Y);
                q = hsv.Z * (1 - hsv.Y * f);
                t = hsv.Z * (1 - hsv.Y * (1 - f));
                if (i == 0)
                {
                    rgb.X = hsv.Z;
                    rgb.Y = t;
                    rgb.Z = p;
                }
                else if (i == 1)
                {
                    rgb.X = q;
                    rgb.Y = hsv.Z;
                    rgb.Z = p;
                }
                else if (i == 2)
                {
                    rgb.X = p;
                    rgb.Y = hsv.Z;
                    rgb.Z = t;
                }
                else if (i == 3)
                {
                    rgb.X = p;
                    rgb.Y = q;
                    rgb.Z = hsv.Z;
                }
                else if (i == 4)
                {
                    rgb.X = t;
                    rgb.Y = p;
                    rgb.Z = hsv.Z;
                }
                else
                {
                    rgb.X = hsv.Z;
                    rgb.Y = p;
                    rgb.Z = q;
                }
            }
            return new Color(new Vector4(rgb,A/255f));
        }
    }
}
