//****************************************************************
//
// Gimp XCF Content Pipeline Importer
//
// Written by: Louis Ingenthron
// Last Modified: 28.05.2009
//
//****************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;

// TODO: replace this with the type you want to import.
using TImport = Microsoft.Xna.Framework.Content.Pipeline.Graphics.TextureContent;

namespace XCFProcessorLib
{
    /// <summary>
    /// This class will be instantiated by the XNA Framework Content Pipeline
    /// to import a file from disk into the specified type, TImport.
    /// 
    /// This should be part of a Content Pipeline Extension Library project.
    /// </summary>
    [ContentImporter(".xcf", DisplayName = "XCF Texture - FV Productions", DefaultProcessor = "XCFProcessor")]
    public class XCFImporter : ContentImporter<TImport>
    {
        const int PROPERTY_TYPE_COMPRESSION = 17;
        const int PROPERTY_TYPE_OPACITY = 6;
        const int PROPERTY_TYPE_VISIBLE = 8;
        const int PROPERTY_TYPE_APPLYMASK = 11;
        const int PROPERTY_TYPE_OFFSETS = 15;
        const int PROPERTY_TYPE_BLENDMODE = 7;

        // skip these properties, they are mostly editor states
        int[] NotUsedProperties = { 18, 19, 22, 23, 24, 25, 2, 5, 10, 12, 13, 26, 3, 4, 14, 9, 20, 21, };

        public override TImport Import(string filename, ContentImporterContext context)
        {
            BinaryReader bin = new BinaryReader(File.OpenRead(filename));
            XCFSourceFile source = new XCFSourceFile();

            {
                String FileType = "gimp xcf ";
                char[] arr = bin.ReadChars(FileType.Length);
                for (int i = 0; i < FileType.Length; i++)
                    if (arr[i] != FileType[i])// the file signature isnt even valid!
                        throw new Exception("Error: File is not a valid GIMP XCF file");
            }
            {
                // most of the time it will be "file" because gimp saves as the lowest compatible version
                String[] SupportedVersionTypes = new [] { "file", "v001", "v002" };
                bool okay = false;
                char[] arr = bin.ReadChars(4);
                for (int vt = 0; vt < SupportedVersionTypes.Length; vt++)
                {
                    bool subokay = true;
                    for (int i = 0; i < SupportedVersionTypes[vt].Length; i++)
                        if (arr[i] != SupportedVersionTypes[vt][i])
                            subokay = false;
                    if (subokay)
                        okay = true;
                }
                if(!okay)
                    throw new Exception("Error: XCF version \""+(new String(arr))+"\" is not supported by this Importer");
            }
            if (bin.ReadByte() != 0)//null terminator
                throw new Exception("Error: File Corruption found in XCF file");

            // read in canvas basics
            source.CanvasWidth = bin.ReadUInt32().ReverseEndianness();
            source.CanvasHeight = bin.ReadUInt32().ReverseEndianness();
            source.ColMode = (XCFSourceFile.ColorMode)bin.ReadUInt32().ReverseEndianness();

            // read in canvas properties... ignore most
            while (true)
            {
                uint type = bin.ReadUInt32().ReverseEndianness();
                uint length = bin.ReadUInt32().ReverseEndianness();
                if (type == 0 && length == 0)
                    break;
                bool notUsed = false;
                for (int i = 0; i < NotUsedProperties.Length; i++)
                    if (type == NotUsedProperties[i])
                        notUsed = true;
                if (notUsed)
                {
                    bin.ReadBytes((int)length);
                }
                else if (type == PROPERTY_TYPE_COMPRESSION)
                {
                    source.Compression = (CompressionType)bin.ReadByte();
                }
                else
                    throw new Exception();
            }

            // stores layers as pointers into the file... pain in the ass
            List<uint> layerPointers = new List<uint>();

            while (true)
            {
                uint layerP = bin.ReadUInt32().ReverseEndianness();
                if (layerP == 0)
                    break;
                layerPointers.Add(layerP);
            }

            // stores channels that way too... oh and lots of other stuff... great! ><
            List<uint> channelPointers = new List<uint>();

            while (true)
            {
                uint channelP = bin.ReadUInt32().ReverseEndianness();
                if (channelP == 0)
                    break;
                channelPointers.Add(channelP);
            }

            // lets go through and read the layers
            for (int i = 0; i < layerPointers.Count; i++)
            {
                bin.BaseStream.Position = (long)layerPointers[i];
                Layer l = new Layer();
                // read the basic info
                l.Width = bin.ReadUInt32().ReverseEndianness();
                l.Height = bin.ReadUInt32().ReverseEndianness();
                l.ColMode = (Layer.ColorMode)bin.ReadUInt32().ReverseEndianness();
                l.Name = bin.ReadXCFString();

                // read layer properties, again: ignore most
                while (true)
                {
                    uint type = bin.ReadUInt32().ReverseEndianness();
                    uint length = bin.ReadUInt32().ReverseEndianness();
                    if (type == 0 && length == 0)
                        break;
                    bool notUsed = false;
                    for (int k = 0; k < NotUsedProperties.Length; k++)
                        if (type == NotUsedProperties[k])
                            notUsed = true;
                    if (notUsed)
                    {
                        bin.ReadBytes((int)length);
                    }
                    else if (type == PROPERTY_TYPE_OPACITY)
                    {
                        l.Alpha = (byte)bin.ReadUInt32().ReverseEndianness();
                    }
                    else if (type == PROPERTY_TYPE_VISIBLE)
                    {
                        l.Visible = bin.ReadUInt32().ReverseEndianness() != 0;
                    }
                    else if (type == PROPERTY_TYPE_APPLYMASK)
                    {
                        l.ApplyMask = bin.ReadUInt32().ReverseEndianness() != 0;
                    }
                    else if (type == PROPERTY_TYPE_BLENDMODE)
                    {
                        l.Blending = (Layer.BlendMode)bin.ReadUInt32().ReverseEndianness();
                    }
                    else if (type == PROPERTY_TYPE_OFFSETS)
                    {
                        l.X = bin.ReadInt32().ReverseEndianness();
                        l.Y = bin.ReadInt32().ReverseEndianness();
                    }
                    else
                        bin.ReadBytes((int)length);
                }
                // the hierarchy pointer and mask pointer
                l.hptr = bin.ReadUInt32().ReverseEndianness();
                l.mptr = bin.ReadUInt32().ReverseEndianness();
                source.Layers.Add(l);
            }

            // process layers
            foreach (Layer layer in source.Layers)
            {
                bin.BaseStream.Position = layer.hptr;
                // make the hierarchy, mostly a useless wrapper structure
                Hierarchy h = new Hierarchy();
                layer.Hierarchy = h;
                h.Width = bin.ReadUInt32().ReverseEndianness();
                h.Height = bin.ReadUInt32().ReverseEndianness();
                h.BPP = bin.ReadUInt32().ReverseEndianness();
                if (h.BPP != layer.BytesPerPixel)
                    throw new Exception("Error. BPP does not match up between Layer and Hierarchy. Possible data corruption");
                if (h.Width != layer.Width)
                    throw new Exception("Error. Width does not match up between Layer and Hierarchy. Possible data corruption");
                if (h.Height != layer.Height)
                    throw new Exception("Error. Height does not match up between Layer and Hierarchy. Possible data corruption");
                uint levelPtr = bin.ReadUInt32().ReverseEndianness();
                bin.BaseStream.Position = levelPtr;
                // load in the level, this is where pixel data is stored
                Level l = new Level();
                h.Level = l;
                {
                    l.Width = bin.ReadUInt32().ReverseEndianness();
                    l.Height = bin.ReadUInt32().ReverseEndianness();
                    // create the tile array
                    l.Tiles = new Tile[((l.Width - 1) / 64) + 1][];
                    for (int i = 0; i < l.Tiles.Length; i++)
                        l.Tiles[i] = new Tile[((l.Height - 1) / 64) + 1];

                    // start reading in tiles
                    bin.BaseStream.Position = bin.ReadUInt32().ReverseEndianness();
                    for (int y = 0; y < ((l.Height-1)/64)+1; y++)
                        for (int x = 0; x < ((l.Width-1)/64)+1; x++)
                        {
                            if(source.Compression == CompressionType.None)
                                l.Tiles[x][y] = Tile.Read(bin, layer.ColMode, (int)((l.Width - (x * 64)) >= 64 ? 64 : (l.Width - (x * 64))), (int)((l.Height - (y * 64)) >= 64 ? 64 : (l.Height - (y * 64))));
                            else
                                l.Tiles[x][y] = Tile.ReadRLE(bin, layer.ColMode, (int)((l.Width - (x * 64)) >= 64 ? 64 : (l.Width - (x * 64))), (int)((l.Height - (y * 64)) >= 64 ? 64 : (l.Height - (y * 64))));
                        }
                }
            }

            bin.Close();

            // put the data in the correct types so the XNA Texture Processor can do its magic
            Texture2DContent tContent = new Texture2DContent();
            MipmapChain chain = new MipmapChain();
            BitmapContent bmpContent = new XCFContent(source);
            chain.Add(bmpContent);
            tContent.Faces[0] = chain;
            tContent.Validate();
            return tContent;
        }
    }
}
